from .Require import require
from .Ensure import ensure
from .Invariant import invariant
from .Types import types
from .Preserve import preserve

__all__ = ["require", "invariant", "preserve", "ensure", "types"]


from .Contract import *
from inspect import cleandoc

class require(Contract):
    def checkPreCondition(self):
        return self.condition(self.args)

    def makePreConditionErrorString(self):
        string = f"""
            Requirement failed at {self.getContractFileLocation()}
            {self.args}
            {self.makeFormattedDescription()}
        """
        return cleandoc(string)

    def makeFormattedDescription(self):
        return super().makeFormattedDescription(args = self.args)


from .Common import *
from functools import wraps

PRESERVER_ATTRIBUTE = '__contract_preserved_values__'

class Contract:
    def __init__(self, condition, description = ""):
        self.callerFrame = getCallerData()
        self.condition = condition
        self.description = description
        self.func = None
        self.args = None
        self.preserved = None
        self.result = None

    def __call__(self, func):
        self.func = func
        @wraps(func)
        def inner(*args, **kwargs):
            self.args = self.getCompleteArguments(*args, **kwargs)

            self.preserved = self.preserveArguments(getattr(inner, PRESERVER_ATTRIBUTE, []))

            if not self.checkPreCondition():
                self.preConditionFailure()

            self.result = self.runFunction()

            if not self.checkPostCondition():
                self.postConditionFailure()

            return self.result

        return inner

    def getCompleteArguments(self, *args, **kwargs):
        completeArgs = makeCompleteArgumentDict(self.func, args, kwargs)
        completeArgs = dictToNamedTuple(completeArgs, "Args")
        return completeArgs

    def preserveArguments(self, preservers):
        return dictToNamedTuple({}, "Old")

    def checkPreCondition(self):
        return True

    def preConditionFailure(self):
        raise PreConditionError(self.makePreConditionErrorString())

    def makePreConditionErrorString(self):
        return "PreCondition Failure"

    def runFunction(self):
        return self.func(*self.args)

    def checkPostCondition(self):
        return True

    def postConditionFailure(self):
        raise PostConditionError(self.makePostConditionErrorString())

    def makePostConditionErrorString(self):
        return "PostCondition Failure"

    def getContractFileLocation(self):
        return makeFileLocationString(self.callerFrame.filename, self.callerFrame.lineno)

    #called _self so self can be passed in without conflict
    def makeFormattedDescription(_self, **kwargs):
        try:
            return _self.description.format(**kwargs)
        except Exception as e:
            return "Description formatting failed: " + str(e)

class PreConditionError(AssertionError):
    pass

class PostConditionError(AssertionError):
    pass


from .Contract import *

def passPreservers(inner, func):
    setattr(inner, PRESERVER_ATTRIBUTE, getattr(func, PRESERVER_ATTRIBUTE, []))

class preserve(Contract):
    def __init__(self, preserver):
        super().__init__(lambda : True)
        self.preserver = preserver

    def __call__(self, func):
        inner = super().__call__(func)
        if not hasattr(func, PRESERVER_ATTRIBUTE):
            setattr(func, PRESERVER_ATTRIBUTE, [])
        getattr(func, PRESERVER_ATTRIBUTE).append(self.preserver)
        passPreservers(inner, func)
        return inner


from .Common import *
from .Contract import Contract
from inspect import cleandoc

class ensure(Contract):
    def checkPostCondition(self):
        if self.condition.__code__.co_argcount == 3:
            return self.condition(self.args, self.result, self.preserved)
        return self.condition(self.args, self.result)

    def makePostConditionErrorString(self):
        string = f"""
                    Ensure failed at {self.getContractFileLocation()}
                    {self.args}
                    {self.preserved}
                    Result = {self.result}
                    {self.makeFormattedDescription()}
                """
        return cleandoc(string)

    def makeFormattedDescription(self):
        return super().makeFormattedDescription(args = self.args, old = self.preserved, result = self.result)

    def preserveArguments(self, preservers):
        x = preserveValues(preservers, self.args)
        return dictToNamedTuple(x, "Old")


from inspect import signature
from collections import namedtuple
from inspect import stack

def dictToNamedTuple(d: dict, name: str):
    return namedtuple(name, d.keys())(**d)

def makeCompleteArgumentDict(func, args, kwargs):
    s = signature(func)
    p = s.parameters
    build = {}
    for a, b in zip(args, p.keys()):
        build[b] = a
    for k, v in p.items():
        if v.default:
            if k not in build:
                build[k] = v.default
    for k, v in kwargs.items():
        build[k] = v

    return build

def makeFileLocationString(filename: str, lineNum):
    return f'File "{filename}", line {lineNum}'

def getCallerData():
    #third element because it gets the caller's caller's info
    return stack()[2]

def preserveValues(preservers, Args):
    preserved = {}
    for preserver in preservers:
        preserved.update(preserver(Args))
    return preserved


from .Common import *
from .Contract import Contract
from inspect import cleandoc

def getArgumentTypeMismatch(args, requirments):
    mismatches = {}
    for (k, v) in requirments.items():
        if k == 'RESULT':
            pass
        elif type(args.__getattribute__(k)) != v:
            mismatches[k] = (args.__getattribute__(k), v)
    return mismatches

class types(Contract):
    def __init__(self, **requirements):
        super().__init__(None)
        self.callerFrame = getCallerData()
        self.requirements = requirements
        self.mismatches = None

    def checkPreCondition(self):
        self.mismatches = getArgumentTypeMismatch(self.args, self.requirements)
        return len(self.mismatches) == 0

    def checkPostCondition(self):
        if 'RESULT' in self.requirements:
            return type(self.result) == self.requirements['RESULT']
        return True

    def makePreConditionErrorString(self):
        string = f"""
            Type check failed at {self.getContractFileLocation()}
            {self.args}
            {self.makeMismatchErrorString()}
            {self.makeFormattedDescription(args = self.args)}
        """
        return cleandoc(string)

    def makePostConditionErrorString(self):
        string = f"""
            Type check failed at {self.getContractFileLocation()}
            {self.args}
            {self.preserved}
            Result = {self.result}, type {type(self.result)}.  Expected type {self.requirements['RESULT']}
            {self.makeFormattedDescription(args = self.args, old = self.preserved, result = self.result)}
        """
        return cleandoc(string)

    def makeMismatchErrorString(self):
        return '\n'.join(f"{k} = {v}, which is type {type(v)}, expected type {t}"
                         for k, (v, t) in self.mismatches.items())


from setuptools import setup, find_packages

setup(name='fracture-contracts',
      version='0.1',
      description='dpcontracts with improved error reporting',
      url='https://github.com/Fracture17/contracts',
      author='Fracture',
      packages=find_packages(),
      install_requires=[
          'hypothesis',
          'pytest',
          'pytest-pycharm'
      ])


from .Common import *
from .Contract import Contract
from inspect import cleandoc, getsourcelines, ismethod, isfunction

#Taken from dpcontracts
def shouldApplyInvariant(name, func, cls):
    exceptions = ("__getitem__", "__setitem__", "__lt__", "__le__", "__eq__",
                  "__ne__", "__gt__", "__ge__", "__init__")

    if name.startswith("__") and name.endswith("__") and name not in exceptions:
        return False

    if not ismethod(func) and not isfunction(func):
        return False

    if getattr(func, "__self__", None) is cls:
        return False

    return True


def getMethodActualFirstLine(func):
    sourceLines = getsourcelines(func)
    line = sourceLines[1]
    sourceLines = sourceLines[0]
    while not sourceLines[0].strip().startswith('def'):
        sourceLines = sourceLines[1:]
        line += 1
    return line


class InvariantCheck(Contract):
    def __init__(self, condition, description, callerFrame):
        super().__init__(condition, description)
        self.callerFrame = callerFrame

    def checkPreCondition(self):
        if self.func.__name__ != '__init__':
            return self.condition(self.args[0])
        return True

    def checkPostCondition(self):
        return self.condition(self.args[0])

    def makePreConditionErrorString(self):
        string = f"""
            PreCheck Invariant failed at {self.getContractFileLocation()}
            During method at {self.getMethodFileLocation()}
            {self.args}
            {self.makeFormattedDescription(self = self.args[0], old = None, result = None)}
        """
        return cleandoc(string)

    def makePostConditionErrorString(self):
        string = f"""
            PreCheck Invariant failed at {self.getContractFileLocation()}
            During method at {self.getMethodFileLocation()}
            {self.args}
            {self.preserved}
            Result = {self.result}
            {self.makeFormattedDescription(self = self.args[0], old = self.preserved, result = self.result)}
        """
        return cleandoc(string)

    def getMethodFileLocation(self):
        methodLine = getMethodActualFirstLine(self.func)
        return makeFileLocationString(self.callerFrame.filename, methodLine)

    def preserveArguments(self, preservers):
        x = preserveValues(preservers, self.args)
        return dictToNamedTuple(x, "Old")

class invariant(Contract):
    def __call__(self, cls):
        class InvariantContractor(cls):
            def __repr__(self):
                return repr(cls)

        for name, value in [(name, getattr(cls, name)) for name in dir(cls)]:
            if shouldApplyInvariant(name, value, cls):
                setattr(InvariantContractor, name,
                        InvariantCheck(self.condition, self.description, self.callerFrame)(value))

        return InvariantContractor